package ca.yorku.cse.mack.sak;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Typeface;
import android.util.AttributeSet;
import android.util.Log;
import android.widget.TextView;

public class QueryPanel extends TextView
{
	// final int PANEL_WIDTH = 730;
	// final int PANEL_HEIGHT = 150;
	final int HIGHLIGHT_COLOR = Color.RED;
	final int NORMAL_COLOR = Color.BLACK;

	Paint p;
	String[] query = { "" };
	FormattedQuery[] formattedQuery;
	int idx = 0; // index of currently active word in query panel

	TextView[] key; // a reference to a key (to get font size)
	float textSize;
	float pixelDensity;

	public QueryPanel(Context contextArg)
	{
		super(contextArg);
		initialize(contextArg);
	}

	public QueryPanel(Context contextArg, AttributeSet attrs)
	{
		super(contextArg, attrs);
		initialize(contextArg);
	}

	public QueryPanel(Context contextArg, AttributeSet attrs, int defStyle)
	{
		super(contextArg, attrs, defStyle);
		initialize(contextArg);
	}

	private void initialize(Context c)
	{
		p = new Paint();
		pixelDensity = c.getResources().getDisplayMetrics().density;
	}

	public void setKeyReference(TextView[] keyArg)
	{
		key = keyArg;
	}

	public void onDraw(Canvas c)
	{
		for (FormattedQuery fq : formattedQuery)
		{
			if (fq.highlight == FormattedQuery.HIGHLIGHT)
			{
				p.setColor(HIGHLIGHT_COLOR);
				p.setTypeface(Typeface.defaultFromStyle(Typeface.BOLD));
			} else
			{
				p.setColor(NORMAL_COLOR);
				p.setTypeface(Typeface.defaultFromStyle(Typeface.NORMAL));
			}
			p.setTextSize(textSize * 1.5f * pixelDensity); // add "*2f" for Nexus 4
			c.drawText(fq.word, fq.x, fq.y, p);
		}
	}

	public void updateQueryPanel(String[] queryArg)
	{
		query = queryArg;
		formattedQuery = new FormattedQuery[query.length];

		int x = 5;
		int h = -(int)p.ascent() + 3; // ascent is negative (negate)
		int y = h;

		for (int i = 0; i < query.length; ++i)
		{
			String s = query[i] + "  ";

			if (x + p.measureText(s) > this.getWidth()) // start new line
			{
				x = 5;
				y += h;
			}
			formattedQuery[i] = new FormattedQuery(s, x, y, FormattedQuery.NORMAL);
			x += p.measureText(s);
		}
		invalidate();
	}

	public void startHighlight()
	{
		idx = 0;
		formattedQuery[idx].highlight = FormattedQuery.HIGHLIGHT;
		invalidate();
	}

	public void advanceHighlight()
	{
		formattedQuery[idx].highlight = FormattedQuery.NORMAL;
		++idx;
		if (idx == formattedQuery.length)
			idx = 0;
		formattedQuery[idx].highlight = FormattedQuery.HIGHLIGHT;
		invalidate();
	}

	public String getHighlightedWord()
	{
		return formattedQuery[idx].word.trim();
	}

	public int getIdx()
	{
		return idx;
	}

	public void clear()
	{
		idx = 0;
		String[] blank = { "" };
		updateQueryPanel(blank);
	}

	@Override
	public void onSizeChanged(int wArg, int hArg, int wOldArg, int hOldArg)
	{		
		//Need to set the text size here!		 
		textSize = key[0].getTextSize();
		
		float keyWidth = key[0].getWidth();
		Log.i("DEBUG", "textSize=" + textSize + ", keyWidth=" + keyWidth);
		while (true)
		{
			float textWidth = key[0].getPaint().measureText("QRSTUVWXYZ");
			if (textWidth < 0.95f * keyWidth)
				break;
			key[0].setTextSize(--textSize);
		}

		// adjust for the other keys, as well
		key[1].setTextSize(textSize);
		key[2].setTextSize(textSize);
		key[3].setTextSize(textSize);
		//this.setTextSize(textSize); // doesn't work (see onDraw)
		updateQueryPanel(query);
	}

	class FormattedQuery
	{
		final static int NORMAL = 1;
		final static int HIGHLIGHT = 2;
		String word;
		int x, y, highlight;

		FormattedQuery(String wordArg, int xArg, int yArg, int highlightArg)
		{
			word = wordArg;
			x = xArg;
			y = yArg;
			highlight = highlightArg;
		}
	}
}
