package ca.yorku.cse.mack.sak;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.graphics.Color;
import android.media.MediaPlayer;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.os.Environment;
import android.os.Vibrator;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnTouchListener;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.TextView;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Locale;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * <h1>SAK</h1>
 *
 * <h3>Summary</h3>
 *
 * <ul>
 * <li>Android experiment software investigating text entry using a Scanning Ambiguous Keyboard (SAK)
 * </ul>
 * <p>
 *
 * <h3>Related references</h3>
 * <p>
 *
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/assets2009.html"> The one-key challenge: Searching for a fast
 * one-key text entry method </a>, by MacKenzie (<i>ASSETS 2009</i>).
 * <p>
 *
 * <li>
 * <a href="http://www.yorku.ca/mack/TOCHI2010.html"> SAK: Scanning ambiguous keyboard for efficient
 * one-key text entry </a>, by MacKenzie and Felzer (<i>TOCHI 2010</i>).
 * <p>
 *
 * <li>
 * <a href="http://www.yorku.ca/mack/UAIS2011.html"> BlinkWrite: Efficient text entry using eye
 * blinks </a>, by MacKenzie and Ashtiani (<I>UAIS 2010</I>).
 * <p>
 *
 * <li>
 * <a href="http://www.yorku.ca/mack/icchp2010b.html"> Qanti: A software tool for quick ambiguous
 * non-standard text input </a>, by Felzer, MacKenzie, Beckerle, and Rinderknecht (<i>ICCHP
 * 2010</i>).
 * <p>
 *
 * <li>
 * <a href="http://www.yorku.ca/mack/ETRA2010.html"> BlinkWrite2: An improved text entry method
 * using eye blinks </a>, by Ashtiani and MacKenzie (<i>ETRA 2010</i>).
 * <p>
 * </ul>
 * <p>
 *
 * <h3>Setup Parameters</h3>
 *
 * The application launches with a setup dialog: (click to enlarge)<p>
 *
 * <center> <a href="SAK-1.jpg"><img src="SAK-1.jpg" width="200" alt="image"></a> </center> <p>
 *
 * In all, there are seven setup parameters, organized in six spinners and one text field. The setup parameters are
 * explained as follows. <p>
 *
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" valign="top">
 *
 * <tr bgcolor="#cccccc">
 * <th align="center">Parameter
 * <th align="center">Description
 *
 * <tr valign="top"> <td>Participant Code <td>A code to identify the user being tested. In experimental research, a user
 * is typically called a "participant". <p>
 *
 * <tr valign="top"> <td>Session Code <td>A code to identify the session. This code could be used in an a longitudinal
 * experiment, where participants are tested over multiple sessions. Each session might involve an hour of testing, with
 * one session per day over multiple days. <p>
 *
 * <tr valign="top">
 * <td>Block Code
 * <td>Note: There is an inactive setup parameter for "Block code". The block code is generated automatically by the
 * software. <p>
 *
 * <tr valign="top"> <td>Group Code <td>A code to identify the group. If the experiment uses a within-subjects design,
 * participants are usually assigned to groups. This code identifies the group. This procedure is known as
 * <i>counterbalancing</i>, which is used to offset order effects. (See section 5.11 in the course's Suggested Readings
 * for complete details.) <p>
 *
 * <tr valign="top"> <td>Condition Code <td>A code to identify an external condition. This setup
 * item is an
 * editable
 * text field. Tap to popup an input method editor to change the value. <p>
 *
 * The condition code is useful if the experiment included an independent variable pertaining to an environmental
 * condition or a context of use. As an example, an experiment might include "mobility" as an independent variable with
 * levels "sitting" and "walking". The code "Sitting" might be used for the sitting condition and the code "Walking" for
 * the walking condition. These codes appear in the output data to facilitate analyses. BTW, an example of an experiment
 * using this independent variable is described by Fitton et al. in <a href="http://www.yorku.ca/mack/bhci2013
 * .html">Exploring Tilt-Based Text Input For Mobile Devices With Teenagers</a>. <p>
 *
 * <tr valign="top"> <td>Phrases per Block<td>The number of phrases entered for each block of
 * testing.
 * <p>
 *
 * <tr valign="top"> <td>Scanning Interval <td>A value between 100 ms and 1500 ms, in 100 ms increments.  This value
 * set the scanning interval &ndash; the time each key maintains focus during the key-to-key scanning process.
 * </table> </blockquote> <p>
 *
 * <h3>Program Operation</h3>
 *
 * After the setup parameters are set, the program is launched by clicking the OK button.  The following screen appears:
 * <p>
 *
 * <center>
 * <a href="SAK-2.jpg"><img src="SAK-2.jpg" width=200 alt="image"></a>
 * </center>
 * <p>
 *
 * A phrase of text appears for input.  Scanning takes place across the three alpha keys and the [word] key seen in the
 * middle of the screen image.  To enter a letter, the user taps on the indicated region when the key bearing the
 * desired letter is highlighted.  Only one tap is required for each letter.  A key may be tapped more than once in a
 * single scanning interval if more than one consecutive letter are on the same key.  In this case, the scan interval
 * is extended by resetting the timer with each tap.<p>
 *
 * As tapping proceeds, a list of candidate
 * words is presented below the letter keys.  After entering all the letters for a word (or earlier for common words)
 * , the user taps the [word] key when highlighted.  Scanning proceeds in the word region.  When the desired word is
 * highlighted, the user taps again.  This enters the word in the transcribed text field followed by a SPACE character.
 * <p>
 *
 * At the end of the phrase, popup results appear.  An example is shown below.
 * <p>
 *
 * <center>
 * <a href="SAK-3.jpg"><img src="SAK-3.jpg" width=200 alt="image"></a>
 * </center>
 * <p>
 *
 * <h3>Data Storage</h3>
 *
 * Data are also saved in files for follow-up analyses.  There are two files, an "sd1" file which contains data on a
 * per-tap basis, and an "sd2" file which contains summary data for each phrase of input.  The data files are
 * saved in
 * the public directory on Android devices in a sub-directory called "SAKData".  Below are example
 * files:
 * <p>
 *
 * <ul>
 * <li>sd1 &ndash; <a href="SAK-P01-S01-B01-G01-C01-SI800.sd1">SAK-P01-S01-B01-G01-C01-SI800.sd1</a>
 * <li>sd2 &ndash; <a href="SAK-P01-S01-B01-G01-C01-SI800.sd2">SAK-P01-S01-B01-G01-C01-SI800.sd2</a>
 * </ul>
 * <p>
 *
 * The data in the sd2 files are comma-delimited to facilitate importing in Excel or a statistics
 * application for follow-up analyses.  An example is shown below for the data in the sd2 file above.
 * <p>
 *
 * <center>
 * <a href="SAK-4.jpg"><img src="SAK-4.jpg" width=800 alt="image"></a>
 * </center>
 *
 * @author Scott MacKenzie, 2011-2017
 */
public class SAKActivity extends Activity implements OnTouchListener
{
    private final static String MYDEBUG = "MYDEBUG"; // for Log.i messages

    private final static String APP = "SAK";
    private final static String DATA_DIRECTORY = "/SAKData/";

    private final static String SD1_HEADER = "scanTimer,mode,key_num,key_code,selected,number_of_selections\n";
    private final static String SD2_HEADER = "App,Participant,Session,Block,Group,Condition,ScanInterval," +
            "PresentedText,TranscribedText,MinScans,Time(s),Scans,Selections,EntrySpeed(wpm),ErrorRate(%),Efficiency" +
            "(%),MultipleSelectCount,MultipleSelectOpportunities,EarlyWordSelect,EarlyWordSelectOpportunities," +
            "WordRestarts,WordDeletes\n";
    private final static int LETTER = 0;
    private final static int WORD = 1;
    private final static int HOLD_DURATION = 1500;

    String sd2Leader;
    private String participantCode, sessionCode, blockCode, groupCode, conditionCode;
    private int phrasesPerBlock, scanInterval;
    private BufferedWriter sd1, sd2;
    private File f1, f2;
    private EncodedWord[] ew;
    private CountDownTimer scanTimer, touchAndHoldTimer;
    private boolean firstWord, endOfBlock, endOfPhrase;
    private int keyPosition;
    private TextView[] key;
    private QueryPanel qp;
    private String[] phrases;
    private EditText presentedTextField, transcribedTextField, currentKeysField;
    private String presentedText, transcribedText, currentKeys;
    private int mode, selections;
    private int scanCount;
    private long elapsedTime, startTime;
    private boolean begin, wordJustSelected;
    private StringBuilder sd1Data, sd2Data;
    private String wordSave;
    private MediaPlayer tapSound;
    private MediaPlayer endOfPhraseSound;
    private MediaPlayer deleteSound;
    private Vibrator vib;
    private int idxSave = 0;
    private int phraseCount = 0;
    private int wordDeletes = 0;
    private int wordRestarts = 0;
    private int multipleSelectCount = 0; // count number of MLPK opportunities taken
    private int earlyWordSelect = 0; // count number of early word selections
    private int totalSelections = 0;

    @Override
    public void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);

        Bundle b = getIntent().getExtras();
        participantCode = b.getString("participantCode");
        sessionCode = b.getString("sessionCode");
        // blockCode (auto)
        groupCode = b.getString("groupCode");
        conditionCode = b.getString("conditionCode");
        phrasesPerBlock = b.getInt("numberOfPhrases");
        scanInterval = b.getInt("scanInterval");
        currentKeys = "";

        presentedTextField = (EditText)findViewById(R.id.presented);
        transcribedTextField = (EditText)findViewById(R.id.transcribed);
        currentKeysField = (EditText)findViewById(R.id.keys);

        tapSound = MediaPlayer.create(getApplicationContext(), R.raw.hit);
        endOfPhraseSound = MediaPlayer.create(getApplicationContext(), R.raw.blip1);
        deleteSound = MediaPlayer.create(getApplicationContext(), R.raw.miss);
        vib = (Vibrator)getSystemService(Context.VIBRATOR_SERVICE);

        mode = LETTER;
        scanCount = selections = 0;
        elapsedTime = 0;
        sd2Data = new StringBuilder(1000);

        // define keyboard
        key = new TextView[4];
        key[0] = (TextView)findViewById(R.id.button0);
        key[1] = (TextView)findViewById(R.id.button1);
        key[2] = (TextView)findViewById(R.id.button2);
        key[3] = (TextView)findViewById(R.id.button3);

        qp = (QueryPanel)findViewById(R.id.query);

		/*
         * Give the query panel a reference to the key array. This will be used in the query panel's
		 * onSizeChanged method to set the text size for the words displayed in the query panel.
		 */
        qp.setKeyReference(key);

        TextView spaceBar = (TextView)findViewById(R.id.spacebar);
        spaceBar.setOnTouchListener(this);

        // ===================
        // File initialization
        // ===================

        // make a working directory (if necessary) to store data files
        File dataDirectory = new File(Environment.getExternalStorageDirectory() +
                DATA_DIRECTORY);
        if (!dataDirectory.exists() && !dataDirectory.mkdirs())
        {
            Log.e(MYDEBUG, "ERROR --> FAILED TO CREATE DIRECTORY: " + DATA_DIRECTORY);
            this.finish(); // terminate
        }

        /**
         * The following do-loop creates data files for output and a string sd2Leader to write to the sd2
         * output files.  Both the filenames and the sd2Leader are constructed by combining the setup parameters
         * so that the filenames and sd2Leader are unique and also reveal the conditions used for the block of input.
         *
         * The block code begins "B01" and is incremented on each loop iteration until an available
         * filename is found.  The goal, of course, is to ensure data files are not inadvertently overwritten.
         */
        int blockNumber = 0;
        do
        {
            ++blockNumber;
            blockCode = String.format(Locale.CANADA, "B%02d", blockNumber);
            String baseFilename = String.format(Locale.CANADA, "%s-%s-%s-%s-%s-%s-SI%d", APP, participantCode,
                    sessionCode, blockCode, groupCode, conditionCode, scanInterval);
            f1 = new File(dataDirectory, baseFilename + ".sd1");
            f2 = new File(dataDirectory, baseFilename + ".sd2");

            // also make a comma-delimited leader that will begin each data line written to the sd2 file
            sd2Leader = String.format(Locale.CANADA, "%s,%s,%s,%s,%s,%s,SI%d", APP, participantCode, sessionCode,
                    blockCode, groupCode, conditionCode, scanInterval);
        } while (f1.exists() || f2.exists());

        try
        {
            sd1 = new BufferedWriter(new FileWriter(f1));
            sd2 = new BufferedWriter(new FileWriter(f2));

            // output header in sd1 file
            sd1.write(SD1_HEADER, 0, SD1_HEADER.length());
            sd1.flush();

            // output header in sd2 file
            sd2.write(SD2_HEADER, 0, SD2_HEADER.length());
            sd2.flush();

        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR OPENING DATA FILES! e=" + e.toString());
            super.onDestroy();
            this.finish();

        } // end file initialization

        // prevent soft keyboard from popping up
        getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

        phrases = getResources().getStringArray(R.array.phrases2);
        presentedText = getNextPhrase(phrases);
        presentedTextField.setText(presentedText);
        transcribedText = "";

        // initialize the EncodedWord array (ew)
        String[] s = getResources().getStringArray(R.array.wordFreq);
        StringTokenizer st;
        Vector<String> w = new Vector<String>();
        Vector<String> f = new Vector<String>();
        for (String value : s)
        {
            st = new StringTokenizer(value);
            w.add(st.nextToken());
            f.add(st.nextToken());
        }
        String[] wArray = new String[w.size()];
        String[] fArray = new String[f.size()];
        w.copyInto(wArray);
        f.copyInto(fArray);
        ew = new EncodedWord[wArray.length];
        for (int i = 0; i < ew.length; ++i)
        {
            String word = wArray[i].toLowerCase();
            int freq = Integer.parseInt(fArray[i]);
            String keystrokes = "";
            for (int j = 0; j < word.length(); ++j)
            {
                // scan alpha keys to look for match (last key is [space])
                for (int k = 0; k < key.length - 1; ++k)
                {
                    String tmp1 = key[k].getText().toString().toLowerCase();
                    String tmp2 = word.substring(j, j + 1);
                    if (tmp1.contains(tmp2)) // found key containing letter
                    {
                        keystrokes += k + 1;
                        break;
                    }
                }
            }
            ew[i] = new EncodedWord(word, freq, keystrokes);
        }
        Arrays.sort(ew, new ByCode());

        // this is the main timer for scanning
        scanTimer = new CountDownTimer(scanInterval, scanInterval)
        {
            public void onTick(long millisUntilFinished)
            {
            }

            public void onFinish()
            {
                doEndOfDwellTime();
            }
        };
        scanTimer.start();

        // this is the timer for touch-and-hold (used for deleting text)
        touchAndHoldTimer = new CountDownTimer(HOLD_DURATION, HOLD_DURATION)
        {
            public void onTick(long millisUntilFinished)
            {
            }

            public void onFinish()
            {
                doHoldTimeout();
            }
        }; // will be started on each touch (MotionEvent.ACTION_DOWN)

    } // end onCreate

    @Override
    public boolean onTouch(View v, MotionEvent me)
    {
        if (me.getAction() == MotionEvent.ACTION_UP)
        {
            touchAndHoldTimer.cancel();
        } else if (me.getAction() == MotionEvent.ACTION_DOWN)
        {
            // finish the app on the first touch after the final results dialog is dismissed
            if (endOfBlock)
            {
                Log.i(MYDEBUG, "onTouch: endOfBlock test");
                tapSound.release();
                endOfPhraseSound.release();
                deleteSound.release();
                finish();
                return true;
            }

            if (!begin)
            {
                begin = true;
                startTime = System.currentTimeMillis();
            }
            ++selections; // per scan step
            ++totalSelections; // per phrase
            scanTimer.cancel();
            scanTimer.start(); // timer restart on select (part of SAK concept)

			/*
             * Start the hold timer. During "normal" operation, the timeout will never occur because the timer is
             * cancelled upon ACTION_UP (see above). However, if the user continues to touch the input key, then the
             * hold timeout will occur. In the doHoldTimeout method (see above), we delete either current keystrokes
             * (if in the middle of a word) or the last word (if between words).
			 */
            touchAndHoldTimer.start();

            if (mode == LETTER)
            {
                if (keyPosition != 3) // not SPACE key
                {
                    currentKeys += "" + (keyPosition + 1);
                    currentKeysField.setText(currentKeys);
                    String[] query = getCandidates(currentKeys, ew);

                    // Are there any candidates for the current key sequence?
                    if (query == null) // NO
                    {
                        vib.vibrate(10);
                        return true; // do nothing (user must select from current query)
                    } else
                    // We're good. Update the panel with the new word choices
                    {
                        tapSound.start();
                        qp.updateQueryPanel(query);
                    }
                } else
                // SPACE key
                {
                    tapSound.start();

                    // is there a word in progress?
                    if (currentKeys.length() > 0) // YES
                    {
                        // switch to word mode to select a word from the query list
                        mode = WORD;
                        firstWord = true;
                    }
                }
            } else if (mode == WORD) // a word is selected
            {
                tapSound.start();
                wordJustSelected = true;
                String word = qp.getHighlightedWord();
                wordSave = word;
                if ((currentKeys.length()) < word.length())
                    ++earlyWordSelect;
                idxSave = qp.getIdx();
                transcribedText += word + " ";
                transcribedTextField.setText(transcribedText);
                transcribedTextField.setSelection(transcribedTextField.getText().length()); // move cursor to end

                // prepare for next word
                mode = LETTER;
                currentKeys = "";
                currentKeysField.setText(currentKeys);
                qp.clear();

                // Is this the end of the phrase?
                if (transcribedText.trim().length() >= presentedText.length()) // YES
                {
                    endOfPhrase = true;
                    endOfPhraseSound.start();
                    transcribedText = transcribedText.trim(); // remove trailing space

                    // gather stats
                    elapsedTime = System.currentTimeMillis() - startTime;
                    float wpm = (presentedText.length() / 5f) / (elapsedTime / 60000f);
                    float er = (float)(new MSD(presentedText, transcribedText)).getErrorRateNew();
                    float efficiency = (float)getMinimumScans(presentedText).length() / scanCount * 100.0f;
                    sd2Data.append(String.format("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s," +
                                    "%s\n", APP, participantCode, sessionCode, blockCode, groupCode, conditionCode,
                            scanInterval, presentedText, transcribedText, getMinimumScans(presentedText),
                            elapsedTime, scanCount, totalSelections, wpm, er, efficiency, multipleSelectCount,
                            getMultipleSelectOpportunities(presentedText), earlyWordSelect,
                            getEarlyWordSelectOpportunities(presentedText), wordRestarts, wordDeletes));

                    /**
                     * NOTE: the elapsed time in the sd2 file is slightly less than the time for
                     * the
                     * last sample in the sd1 file. The sd2 time is more accurate, since it is
                     * coincident with the touch that finished the phrase. The last sd1 sample is
                     * needed, however, to ensure the last word event is recorded.
                     */

                    // reset some stats variables
                    scanCount = 0;
                    multipleSelectCount = 0;
                    earlyWordSelect = 0;
                    wordRestarts = 0;
                    wordDeletes = 0;

                    String resultsString = String.format("End of Phrase.  Thank you.%nPresented text = %s%n"
                            + "Transcribed text = %s%nEntry speed = %.2f wpm%nError rate = %.2f%%%nEfficiency" +
                            " = %.2f%%", presentedText, transcribedText, wpm, er, efficiency);
                    showResultsDialog(resultsString);

                    ++phraseCount;
                    // Is this also the end of the block?
                    if (phraseCount == phrasesPerBlock) // YES (we're done)
                        endOfBlock = true; // will finish at end of current scan
                }
            }
        }
        return true;
    } // end onTouch

    // delete something (either the current keys or the previous word)
    private void doHoldTimeout()
    {
        // Is the user in letter mode?
        if (mode == LETTER) // YES
        {
            // Is the user "between words"?
            if (currentKeys.length() <= 1) // YES
            {
                sd1Data.append("[delete_last_word]\n");
                ++wordDeletes;
                transcribedText = deleteLastWord(transcribedText);
                transcribedTextField.setText(transcribedText);
                // move cursor to end
                transcribedTextField.setSelection(transcribedTextField.getText().length());
            } else
            {
                sd1Data.append("[restart_current_word]\n");
                ++wordRestarts;
            }

            // either way, delete keys currently being entered
            deleteSound.start();
            currentKeys = "";
            currentKeysField.setText(currentKeys);
            qp.clear();
            touchAndHoldTimer.start(); // restart timer (multiple deletes possible in one hold)
        }
    }

    // end of dwell time
    private void doEndOfDwellTime()
    {
        scanTimer.start(); // next scan starts immediately

        // first deal with data collection
        if (begin)
        {
            ++scanCount;
            String modeString;
            int wordIdxOrKeyPosition;
            String wordOrKey;
            if ((mode == LETTER || firstWord) && !wordJustSelected)
            {
                modeString = "letter";
                wordIdxOrKeyPosition = keyPosition;
                wordOrKey = key[keyPosition].getText().toString();
            } else
            {
                modeString = "word";
                if (wordJustSelected)
                {
                    wordIdxOrKeyPosition = idxSave;
                    wordOrKey = wordSave;
                    wordJustSelected = false;
                } else
                {
                    wordIdxOrKeyPosition = qp.getIdx();
                    wordOrKey = qp.getHighlightedWord();
                }
            }
            sd1Data.append(String.format("%s,%s,%s,%s,%s,%s\n", System.currentTimeMillis() -
                    startTime, modeString, wordIdxOrKeyPosition, wordOrKey, selections >
                    0, selections));
        }

        // check if done flag set on last touch (must finish here so last sd1 sample is saved)
        if (endOfBlock)
            finishBlockAndExit();

        if (endOfPhrase)
        {
            endOfPhrase = false;
            finishPhraseAndPrepareForNext();
        }

        // now deal with the GUI
        if (mode == LETTER)
        {
            key[keyPosition].setTextColor(Color.BLACK);
            key[keyPosition].setBackgroundColor(Color.WHITE);
            ++keyPosition;
            if (keyPosition >= key.length)
                keyPosition = 0;
            key[keyPosition].setTextColor(Color.RED);
            key[keyPosition].setBackgroundColor(Color.LTGRAY);
        } else if (mode == WORD)
        {
            if (firstWord)
            {
                qp.startHighlight();
                key[keyPosition].setTextColor(Color.BLACK);
                key[keyPosition].setBackgroundColor(Color.WHITE);
                firstWord = false;
            } else
                qp.advanceHighlight();
        }
        if (selections > 1)
            ++multipleSelectCount;
        selections = 0; // reset selection count at end of each scan interval
    }

    private void showResultsDialog(String text)
    {
        // Instantiate the layout
        Context mContext = getApplicationContext();
        LayoutInflater inflater = (LayoutInflater)mContext.getSystemService
                (LAYOUT_INFLATER_SERVICE);
        View layout = inflater.inflate(R.layout.results_dialog, (ViewGroup)findViewById(R.id
                .results_layout));

        // Set text
        TextView results = (TextView)layout.findViewById(R.id.resultsArea);
        results.setText(text);

        // Initialize the dialog
        AlertDialog.Builder parameters = new AlertDialog.Builder(this);
        parameters.setView(layout).setCancelable(false).setNeutralButton("OK", new
                DialogInterface.OnClickListener()
                {
                    public void onClick(DialogInterface dialog, int id)
                    {
                        scanTimer.start(); // bug fix attempt (27/07/2016)
                        dialog.cancel(); // close this dialog
                    }
                }).show();
    }

    private void finishBlockAndExit()
    {
        writeData(sd1, sd1Data.toString());
        writeData(sd1, transcribedText + "\n#\n");
        writeData(sd2, sd2Data.toString());
        try
        {
            sd1.close();
            sd2.close();

            /**
             * Make the saved data files visible in Windows Explorer. There seems to be a bug doing this with Android
             * 4.4. I'm using the following code, instead of sendBroadcast. See...
             * http://code.google.com/p/android/issues/detail?id=38282
             */
            MediaScannerConnection.scanFile(this, new String[] {f1.getAbsolutePath(),
                    f2.getAbsolutePath()}, null, null);

        } catch (IOException e)
        {
            // work on later
        }

        scanTimer.cancel(); // no need to continue scanning
        // NOTE: Actual exit will occur in onTouch (first touch after last results dialog is dismissed)
    }

    private void finishPhraseAndPrepareForNext()
    {
        writeData(sd1, sd1Data.toString());
        writeData(sd1, transcribedText + "\n#\n");
        presentedText = getNextPhrase(phrases);
        presentedTextField.setText(presentedText);
        transcribedText = "";
        transcribedTextField.setText(transcribedText);
    }

    private String getNextPhrase(String[] phraseArg)
    {
        Random r = new Random();
        int n = r.nextInt(phraseArg.length);

        sd1Data = new StringBuilder(1000); // data collect begins with each phrase
        sd1Data.append(String.format("#\n%s\n", phraseArg[n]));
        begin = false; // first touch after new phrase will change this flag to true
        return phraseArg[n];
    }

    private void writeData(BufferedWriter bw, String s)
    {
        try
        {
            bw.write(s, 0, s.length());
            bw.flush();
        } catch (IOException e)
        {
            // work on later
        }
    }

    private String deleteLastWord(String sArg)
    {
        String s = sArg;
        s = s.trim();
        if (s.length() < 2)
            return s;
        int foundAt = -1;
        for (int i = s.length() - 1; i > 0; --i)
            if (s.charAt(i) == ' ')
            {
                foundAt = i;
                break;
            }
        if (foundAt > 0)
            return s.substring(0, foundAt + 1); // leave trailing space (for now)
        else
            return "";
    }

    private String getMinimumScans(String phraseArg)
    {
        String minScans = "";
        StringTokenizer st = new StringTokenizer(phraseArg);
        while (st.hasMoreTokens())
            minScans += SAKStats.getScanSequence(SAKStats.OW, st.nextToken(), ew);
        int i = 0;
        while (minScans.charAt(i) == '.')
            minScans = minScans.substring(1); // remove scans before 1st selection
        return minScans;
    }

    private int getMultipleSelectOpportunities(String phraseArg)
    {
        int count = 0;
        StringTokenizer st = new StringTokenizer(phraseArg);
        while (st.hasMoreTokens())
            count += SAKStats.getMLPKOpportunities(st.nextToken(), ew);
        return count;
    }

    private int getEarlyWordSelectOpportunities(String phraseArg)
    {
        int count = 0;
        StringTokenizer st = new StringTokenizer(phraseArg.toLowerCase());
        while (st.hasMoreTokens())
            count += SAKStats.getOWOpportunities(st.nextToken(), ew);
        return count;
    }

    private String[] getCandidates(String keycode, EncodedWord[] ew)
    {
        final int MAX = 25; // maximum reasonable list size (before full word entered)

        String[] unique = EncodedWord.getUnique(keycode, ew);
        String[] query = EncodedWord.getQuery(keycode, ew);

        // if nothing happening, return null
        int n1 = unique == null ? 0 : unique.length;
        int n2 = query == null ? 0 : query.length;
        if ((n1 + n2) == 0)
            return null;

        String[] candidates;
        if (n1 >= MAX)
        {
            candidates = new String[MAX];
            if (unique != null)
                System.arraycopy(unique, 0, candidates, 0, MAX);
        } else
        {
            int total = n2 > MAX ? MAX : n2;
            candidates = new String[total];
            if (query != null)
                System.arraycopy(query, 0, candidates, 0, total);
        }
        return candidates;
    }
}