package ca.yorku.cse.mack.sak;

import android.app.Activity;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.view.View;
import android.view.WindowManager;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Spinner;
import android.widget.Toast;

public class SAKSetup extends Activity 
{
	final static String MYEDEBUG = "MYDEBUG";

	// default values for spinner (NOTE: 1st value will be overwritten by default read from shared preference)
	final String[] participantCode = { "P01", "P01", "P02", "P03", "P04", "P05",
		"P06", "P07", "P08", "P09", "P10", "P11", "P12", "P13", "P14", "P15", 
		"P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24", "P25" };		
	final String[] sessionCode = { "S01",
		"S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09", "S10",
		"S11", "S12", "S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20",
		"S21", "S22", "S23", "S24", "S25" };
	String[] blockCode = {"(auto)"};
	String[] groupCode = {"G01", "G01", "G02", "G03", "G04", "G05", "G06", "G07", "G08", "G09", "G10", "G11", "G12",
			"G13", "G14", "G15", "G16", "G17", "G18", "G19", "G20", "G21", "G22", "G23", "G24", "G25"};
	final String[] numberOfPhrases = { "1", "1", "2", "3", "4", "5", "6", "7",
			"8", "9", "10" };
	final String[] scanInterval = { "800", "100", "200", "300", "400", "500", "600",
			"700", "800", "900", "1000", "1100", "1200", "1300", "1400", "1500" };

	// default values for EditText fields (may be different if shared preferences saved)
	String conditionCode = "C01";

	private Spinner spinParticipantCode;
	private Spinner spinSessionCode, spinGroupCode;
	private Spinner spinNumberOfPhrases, spinScanInterval;
	private EditText editConditionCode;
	Button ok, save, exit;

	SharedPreferences sp;
	SharedPreferences.Editor spe;

    // called when the activity is first created
    @Override
    public void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.setup);

		// get a reference to a SharedPreferences object (used to store, retrieve, and save setup parameters)
		sp = this.getPreferences(MODE_PRIVATE);

		// overwrite 1st entry from shared preferences, if corresponding value exits
		participantCode[0] = sp.getString("participantCode", participantCode[0]);
		sessionCode[0] = sp.getString("sessionCode", sessionCode[0]);
		// block code initialized in main activity (based on existing filenames)
		groupCode[0] = sp.getString("groupCode", groupCode[0]);
		conditionCode = sp.getString("conditionCode", conditionCode);
		numberOfPhrases[0] = sp.getString("numberOfPhrases", numberOfPhrases[0]);
		scanInterval[0] = sp.getString("scanInterval", scanInterval[0]);
		
		// get references to widget elements
		spinParticipantCode = (Spinner)findViewById(R.id.spinParticipantCode);
    	spinSessionCode = (Spinner)findViewById(R.id.spinSessionCode);
    	Spinner spinBlockCode = (Spinner)findViewById(R.id.spinBlockCode);
		editConditionCode = (EditText)findViewById(R.id.conditionCode);
		spinGroupCode = (Spinner)findViewById(R.id.spinGroupCode);
    	spinNumberOfPhrases = (Spinner)findViewById(R.id.spinNumberOfPhrases);
    	spinScanInterval = (Spinner)findViewById(R.id.spinScanInterval);

		// get references to OK, SAVE, and EXIT buttons
		ok = (Button)findViewById(R.id.ok);
		save = (Button)findViewById(R.id.save);
		exit = (Button)findViewById(R.id.exit);

    	// initialise spinner adapters
    	ArrayAdapter<CharSequence> adapterPC
    		= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, participantCode);
    	spinParticipantCode.setAdapter(adapterPC);
    	
    	ArrayAdapter<CharSequence> adapterSS
			= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, sessionCode);
    	spinSessionCode.setAdapter(adapterSS);

		ArrayAdapter<CharSequence> adapterB
				= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, blockCode);
		spinBlockCode.setAdapter(adapterB);

		ArrayAdapter<CharSequence> adapterGC
				= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, groupCode);
		spinGroupCode.setAdapter(adapterGC);

		ArrayAdapter<CharSequence> adapterNP
		= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, numberOfPhrases);
    	spinNumberOfPhrases.setAdapter(adapterNP);
    	
    	ArrayAdapter<CharSequence> adapterSI
		= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, scanInterval);
    	spinScanInterval.setAdapter(adapterSI);

		// initialize EditText setup items
		editConditionCode.setText(conditionCode);

		// prevent soft keyboard from popping up when activity launches
		this.getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);
	}

	public void onClick(View v)
	{
		if (v == ok)
		{
			// get user's choices
			String part = participantCode[spinParticipantCode.getSelectedItemPosition()];
			String sess = sessionCode[spinSessionCode.getSelectedItemPosition()];
			// String block = blockCode[spinBlock.getSelectedItemPosition()];
			String group = groupCode[spinGroupCode.getSelectedItemPosition()];
			String cond = editConditionCode.getText().toString();
			int num = Integer.parseInt(numberOfPhrases[spinNumberOfPhrases.getSelectedItemPosition()]);
			int scanInt = Integer.parseInt(scanInterval[spinScanInterval.getSelectedItemPosition()]);

			// package the user's choices in a bundle
			Bundle b = new Bundle();
			b.putString("participantCode", part);
			b.putString("sessionCode", sess);
			// b.putString("blockCode", block);
			b.putString("groupCode", group);
			b.putString("conditionCode", cond);
			b.putInt("numberOfPhrases", num);
			b.putInt("scanInterval", scanInt);

			// start experiment activity (sending the bundle with the user's choices)
			Intent i = new Intent(getApplicationContext(), SAKActivity.class);
			i.putExtras(b);
			startActivity(i);
			finish();

		} else if (v == save)
		{
			spe = sp.edit();
			spe.putString("participantCode", participantCode[spinParticipantCode.getSelectedItemPosition()]);
			spe.putString("sessionCode", sessionCode[spinSessionCode.getSelectedItemPosition()]);
			spe.putString("groupCode", groupCode[spinGroupCode.getSelectedItemPosition()]);
			spe.putString("conditionCode", editConditionCode.getText().toString());
			spe.putString("numberOfPhrases", numberOfPhrases[spinNumberOfPhrases.getSelectedItemPosition()]);
			spe.putString("scanInterval", scanInterval[spinScanInterval.getSelectedItemPosition()]);
			spe.apply();
			Toast.makeText(this, "Preferences saved!", Toast.LENGTH_SHORT).show();

		} else if (v == exit)
		{
			this.finish(); // terminate
		}
	}
}
